from __future__ import annotations

import re
from typing import Any, Dict, List, NamedTuple, Tuple, Type, TypeVar, Union

from inui.config.replaces import replace


def _fix_quoute(string: str):
    string = string.replace("\\", "\\\\")
    if "'" in string and '"' in string:
        string = string.replace('"', '\\"')
        string = string.replace("'", "\\'")
    elif '"' in string:
        string = string.replace('"', "'")
    elif "'" in string:
        string = string.replace("'", '"')
    return string


def _find_quote(string: str):
    if '"' in string and '"' in string:
        string = string.replace('"', '\\"')
        string = string.replace("'", "\\'")
        return f'"{string}"'
    elif "'" in string:
        return f'"{string}"'

    elif '"' in string:
        return f"'{string}'"
    else:
        return f'"{string}"'


class Attributes:
    def __init__(self, **kwargs: Any) -> None:
        self.attrs: Dict[str, Any] = {
            replace(key): value for key, value in kwargs.items()
        }

    # def __getattribute__(self, __name: str) -> Any:
    #     if __name == "attrs":
    #         return super().__getattribute__("attrs")
    #     return self.attrs.get(__name, None)

    def __str__(self) -> str:
        return " ".join(
            f'{replace(name)}="{self._str(value)}"'
            for name, value in self.attrs.items()
            if value is not None
        )

    def fix_values(self, value: str | list = ""):
        if isinstance(value, str):
            return value

    def __getitem__(self, index: str) -> Any:
        return self.attrs.get(index, None)

    def __setitem__(self, index: str, value: Any) -> None:
        self.attrs[index] = value

    def __delitem__(self, index: str) -> None:
        del self.attrs[index]

    def __iter__(self) -> Any:
        return iter(self.attrs)

    def __len__(self) -> int:
        return len(self.attrs)

    def keys(self) -> List[str]:
        return list(self.attrs.keys())

    def values(self) -> List[Any]:
        return list(self.attrs.values())

    def items(self) -> List[Tuple[str, Any]]:
        return list(self.attrs.items())

    def get(self, index: str, default: Any = None) -> Any:
        return self.attrs.get(index, default)

    def update(self, **kwargs: Any) -> None:
        self.attrs.update(kwargs)

    def clear(self) -> None:
        self.attrs.clear()

    def pop(self, index: str, default: Any = None) -> Any:
        return self.attrs.pop(index, default)

    def popitem(self) -> Tuple[str, Any]:
        return self.attrs.popitem()

    def append(self, key: str = "class", *values):
        key = replace(key).lower()
        if key not in self.attrs.keys():
            self.attrs[key] = values
        else:
            if isinstance(self.attrs[key], (list, tuple, set)):
                self.attrs[key] = (*self.attrs[key], *values)
            else:
                self.attrs[key] = (self.attrs[key], *values)

    def _str(self, data, seperator=" ", fill=""):
        try:
            if isinstance(data, type):
                if isinstance(data(), Base):
                    data = data()
        except Exception as _:
            pass

        if isinstance(data, (set, tuple, list)):
            result = seperator.join([self._str(i) for i in data])
            return f"{fill}{result}{fill}"

        if isinstance(data, dict):
            result = ""
            for key, value in data.items():
                result += f"{self._str(key)}: {self._str(value)};"
            return f"{fill}{result}{fill}"
        return str(data)

    def copy(self):
        return self.attrs.copy()


class _Data:
    def __init__(self, *args) -> None:
        self._data: List[Any] = args

    def __str__(self) -> str:
        return "".join(map(self._str, self._data))

    def _str(self, data):
        try:
            if isinstance(data, type):
                if isinstance(data(), Base):
                    data = data()
        except Exception as _:
            pass
        return str(data)

    def append(self, *args):
        self._data = (*self._data, *args)

    def copy(self):
        return self._data.copy()


class _Meta(NamedTuple):
    start_tag: str = None
    end_tag: str = None
    start_sign: str = "<"
    end_sign: str = ">"
    close_sign: str = "</"
    void_sign: str = "/>"


empty = _Meta("", "", "", "", "", "")


class Base:
    _Meta: Type[_Meta] = _Meta()
    _data = _Data()
    _attributes = Attributes()

    def __init__(
        self, *data: Any, _meta: Type[_Meta] | None = None, **attributes: Any
    ) -> None:
        self.data = data
        self.attributes = attributes
        self._Meta = _meta or self._Meta

    @property
    def data(self) -> list:
        return self._data

    @data.setter
    def data(self, value: Any) -> None:
        self._data = _Data(*value)

    @property
    def start_tag(self) -> str:
        return (
            self._Meta.start_tag
            if self._Meta.start_tag is not None
            else self.__class__.__name__.lower()
        )

    @start_tag.setter
    def start_tag(self, value: str) -> None:
        self._Meta.start_tag = value

    @property
    def end_tag(self) -> str:
        return (
            self._Meta.end_tag
            if self._Meta.end_tag is not None
            else self.__class__.__name__.lower()
        )

    @end_tag.setter
    def end_tag(self, value: str) -> None:
        self._Meta.end_tag = value

    @property
    def attributes(self) -> Attributes:
        return self._attributes

    @attributes.setter
    def attributes(self, value: Union[Attributes, Dict[str, Any]]) -> None:
        if not isinstance(value, (Attributes, dict)):
            raise TypeError("attributes must be an instance of Attributes or dict")
        self._attributes = (
            Attributes(**value) if isinstance(value, dict) else value.copy()
        )

    def render_to_string(self, prettify: bool = False) -> str:

        return f"{self._Meta.start_sign}{self.start_tag}{' ' if len(self.attributes) > 0 else ''}\
{str(self.attributes)}{self._Meta.end_sign}{str(self.data)}{self._Meta.close_sign}{self.end_tag}{self._Meta.end_sign}"

    def __str__(self) -> str:
        return self.render_to_string()

    def __repr__(self) -> str:
        return self.render_to_string()

    def __add__(self, other):
        return none(self.copy(), other)

    def __radd__(self, other) -> str:
        return none(other, self.copy())

    def __len__(self):
        return len(self.__str__())

    def __getitem__(self, item):
        return self.__str__()[item]

    def _str(self, data):
        try:
            if isinstance(data, type):
                if isinstance(data(), Base):
                    data = data()
        except Exception as _:
            pass
        return str(data)

    def _compare(self, other, op) -> Base:
        return none(self.copy(), op, other)

    def __lt__(self, other):
        return self._compare(other, "<")

    def __gt__(self, other):
        return self._compare(other, ">")

    def __le__(self, other):
        return self._compare(other, "<=")

    def __ge__(self, other):
        return self._compare(other, ">=")

    def __mul__(self, value: int):
        return none(*[self.copy() for _ in range(int(value))])

    def __rmul__(self, value: int):
        return self.__mul__(value)

    def __imul__(self, value: int):
        return self.__mul__(value)

    def css(self, tag: str = ""):
        return tag + self.start_tag

    def __matmul__(self, other):
        return self.add_attribute(other, key="style")

    def __and__(self, other):
        return self.add_attribute(other, key="id")

    def __or__(self, other):
        return self.add_child(other)

    def __mod__(self, other):
        return self.add_attribute(other, key="class")

    def __truediv__(self, other):
        return self.add_attribute(other, key="type")

    def __floordiv__(self, other):
        return self.add_attribute(other, key="name")

    def __ifloordiv__(self, other):
        return self.add_attribute(other, key="name")

    def __imatmul__(self, other):
        return self.__matmul__(other)

    def __iand__(self, other):
        return self.__and__(other)

    def __ior__(self, other):
        return self.__or__(other)

    def __imod__(self, other):
        return self.__mod__(other)

    def __itruediv__(self, other):
        return self.__truediv__(other)

    def copy(self):
        return self.__class__(*self._data._data, **self._attributes.copy())

    def add_attribute(self, other, key="class"):
        if not isinstance(other, (list, set, tuple)):
            other = (other,)
        self._attributes.append(
            key,
            *other,
        )
        return self

    def add_child(self, other):
        if not isinstance(other, (list, set, tuple)):
            other = (other,)
        self._data.append(
            *other,
        )
        return self

    def from_string(
        self,
        template_name: str = "",
        context: dict[str, Any] | None = None,
        request=None,
    ):
        context = context or {}

        def replace_from_context(match: re.Match):
            var_name = match.group(1)
            default_value = eval(match.group(2))
            r = context.get(var_name, default_value)
            if isinstance(r, str):
                r = _find_quote(r)
            return str(r)

        def replace_request(match: re.Match):
            default_value = eval(match.group(1))
            r = request or default_value
            if isinstance(r, str):
                r = _find_quote(r)
            return str(r)

        template_name = re.sub(
            r"__from_context__\(\s*['\"](\w+)['\"]\s*,\s*(.*?)\s*\)",
            replace_from_context,
            template_name,
        )
        template_name = re.sub(r"__request__\((.*?)\)", replace_request, template_name)

        template_name = re.sub(r"__out__\((.*?)\)", r"output = \1", template_name)
        local_vars = {}
        exec(template_name, {}, local_vars)

        return local_vars.get("output")

    def render(self, template_name, context=None, request=None, base_path: str = ""):
        with open(base_path + template_name) as f:
            text = f.read()
        return self.from_string(text, context, request)


class BaseElement(Base): ...


class BaseVoidElement(BaseElement):
    def render_to_string(self, prettify: bool = False):
        return f"""{self._Meta.start_sign}{self.start_tag}{' ' if len(self.attributes) > 0 else ''}{str(self.attributes)}{self._Meta.close_sign} """


class BaseSvgElement(Base): ...


class Class(Base):
    def __init__(self, *args) -> None:
        self.args = args

    def render_to_string(self, prettify: bool = False) -> str:
        return " ".join(list(map(self._str, self.args)))


class none(Base):
    _Meta = empty


T = TypeVar("T")


def __from_context__(key: str, *defualt: T) -> T:
    return (*defualt,)


def __out__(*args: T) -> T:
    return (*args,)


def __request__(*defualt: T) -> T:
    return (*defualt,)


# if "\n" not in



"""
A utility class for handling various unit conversions.

Conversion Ratios:
- px: pixels
- pt: points
- in: inches
- cm: centimeters
- mm: millimeters
- pc: picas
- em: em units
- rem: root em units

Additional Methods:
- px_to_percent: Convert pixels to percentage based on a container width.
- percent_to_px: Convert percentage to pixels based on a container width.
- cm_to_in: Convert centimeters to inches.
- in_to_cm: Convert inches to centimeters.
"""


def convert(value: float, ratio: float) -> float:
    return value * ratio


def px_to_pt(px: float) -> float:
    return convert(px, 0.75)


def pt_to_px(pt: float) -> float:
    return convert(pt, 1 / 0.75)


def px_to_in(px: float) -> float:
    return convert(px, 0.0138889)


def in_to_px(inch: float) -> float:
    return convert(inch, 1 / 0.0138889)


def px_to_cm(px: float) -> float:
    return convert(px, 0.0352778)


def cm_to_px(cm: float) -> float:
    return convert(cm, 1 / 0.0352778)


def px_to_mm(px: float) -> float:
    return convert(px, 0.352778)


def mm_to_px(mm: float) -> float:
    return convert(mm, 1 / 0.352778)


def px_to_pc(px: float) -> float:
    return convert(px, 0.05)


def pc_to_px(pc: float) -> float:
    return convert(pc, 1 / 0.05)


def px_to_em(px: float) -> float:
    return convert(px, 0.0625)


def em_to_px(em: float) -> float:
    return convert(em, 1 / 0.0625)


def px_to_rem(px: float) -> float:
    return convert(px, 0.0625)


def rem_to_px(rem: float) -> float:
    return convert(rem, 1 / 0.0625)


def px_to_percent(px: float, container_width: float) -> float:
    return (px / container_width) * 100


def percent_to_px(percent: float, container_width: float) -> float:
    return (percent / 100) * container_width


def cm_to_in(cm: float) -> float:
    return cm * 0.393701


def in_to_cm(inch: float) -> float:
    return inch / 0.393701



from typing import Any

from inui.base import BaseVoidElement
from inui.bootstrap.utils import bs_class_to_lower
from inui.conf import FOUNDATION_CSS_URL
from inui.elements import Button, Html, Link


class FoundationEmail(BaseVoidElement):

    def __init__(
        self, *data: Any, url: str = FOUNDATION_CSS_URL, **attributes: Any
    ) -> None:
        self.url = url
        super().__init__(*data, **attributes)

    def render_to_string(self, prettify: bool = False):
        return Link(href=self.url).render_to_string()


class BaseFoundation(BaseVoidElement):
    def __init__(self, *values) -> None:
        self.values = values

    def render_to_string(self, prettify: bool = False) -> str:
        values = ""
        if self.values:
            values += "-"
            values += "-".join(list(map(str, self.values)))
        return bs_class_to_lower(self.__class__.__name__) + values


class button(BaseFoundation): ...


class tiny(BaseFoundation): ...


class small(BaseFoundation): ...


class large(BaseFoundation): ...


class small_expanded(BaseFoundation): ...


class tiny_expanded(BaseFoundation): ...


class large_expanded(BaseFoundation): ...


class colors(BaseFoundation): ...


class secondary: ...


class success(BaseFoundation): ...


class warning(BaseFoundation): ...


class alert(BaseFoundation): ...


Html(
    FoundationEmail,
    Button(
        "salam",
    )
    @ button
    @ tiny,
)



"""
ISO 639-1 Language Codes
ISO 639-1 defines abbreviations for languages:
"""

Abkhazian = "ab"
Afar = "aa"
Afrikaans = "af"
Akan = "ak"
Albanian = "sq"
Amharic = "am"
Arabic = "ar"
Aragonese = "an"
Armenian = "hy"
Assamese = "as"
Avaric = "av"
Avestan = "ae"
Aymara = "ay"
Azerbaijani = "az"
Bambara = "bm"
Bashkir = "ba"
Basque = "eu"
Belarusian = "be"
Bengali_Bangla = "bn"
Bihari = "bh"
Bislama = "bi"
Bosnian = "bs"
Breton = "br"
Bulgarian = "bg"
Burmese = "my"
Catalan = "ca"
Chamorro = "ch"
Chechen = "ce"
Chichewa_Chewa_Nyanja = "ny"
Chinese = "zh"
Chinese_Simplified = "zh-Hans"
Chinese_Traditional = "zh-Hant"
Chuvash = "cv"
Cornish = "kw"
Corsican = "co"
Cree = "cr"
Croatian = "hr"
Czech = "cs"
Danish = "da"
Divehi_Dhivehi_Maldivian = "dv"
Dutch = "nl"
Dzongkha = "dz"
English = "en"
Esperanto = "eo"
Estonian = "et"
Ewe = "ee"
Faroese = "fo"
Fijian = "fj"
Finnish = "fi"
French = "fr"
Fula_Fulah_Pulaar_Pular = "ff"
Galician = "gl"
Gaelic_Scottish = "gd"
Gaelic_Manx = "gv"
Georgian = "ka"
German = "de"
Greek = "el"
Greenlandic = "kl"
Guarani = "gn"
Gujarati = "gu"
Haitian_Creole = "ht"
Hausa = "ha"
Hebrew = "he"
Herero = "hz"
Hindi = "hi"
Hiri_Motu = "ho"
Hungarian = "hu"
Icelandic = "is"
Ido = "io"
Igbo = "ig"
Indonesian = "id in"
Interlingua = "ia"
Interlingue = "ie"
Inuktitut = "iu"
Inupiak = "ik"
Irish = "ga"
Italian = "it"
Japanese = "ja"
Javanese = "jv"
Kalaallisut_Greenlandic = "kl"
Kannada = "kn"
Kanuri = "kr"
Kashmiri = "ks"
Kazakh = "kk"
Khmer = "km"
Kikuyu = "ki"
Kinyarwanda_Rwanda = "rw"
Kirundi = "rn"
Kyrgyz = "ky"
Komi = "kv"
Kongo = "kg"
Korean = "ko"
Kurdish = "ku"
Kwanyama = "kj"
Lao = "lo"
Latin = "la"
Latvian_Lettish = "lv"
Limburgish__Limburger = "li"
Lingala = "ln"
Lithuanian = "lt"
Luga_Katanga = "lu"
Luganda_Ganda = "lg"
Luxembourgish = "lb"
Manx = "gv"
Macedonian = "mk"
Malagasy = "mg"
Malay = "ms"
Malayalam = "ml"
Maltese = "mt"
Maori = "mi"
Marathi = "mr"
Marshallese = "mh"
Moldavian = "mo"
Mongolian = "mn"
Nauru = "na"
Navajo = "nv"
Ndonga = "ng"
Northern_Ndebele = "nd"
Nepali = "ne"
Norwegian = "no"
Norwegian_bokm = "nb"
Norwegian_nynorsk = "nn"
Nuosu = "ii"
Occitan = "oc"
Ojibwe = "oj"
Old_Church_Slavonic_Old_Bulgarian = "cu"
Oriya = "or"
Oromo_Afaan_Oromo = "om"
Ossetian = "os"
Pali = "pi"
Pashto_Pushto = "ps"
Persian = "fa"
Farsi = "fa"
Polish = "pl"
Portuguese = "pt"
Punjabi_Eastern = "pa"
Quechua = "qu"
Romansh = "rm"
Romanian = "ro"
Russian = "ru"
Sami = "se"
Samoan = "sm"
Sango = "sg"
Sanskrit = "sa"
Serbian = "sr"
Serbo_Croatian = "sh"
Sesotho = "st"
Setswana = "tn"
Shona = "sn"
Sichuan_Yi = "ii"
Sindhi = "sd"
Sinhalese = "si"
Siswati = "ss"
Slovak = "sk"
Slovenian = "sl"
Somali = "so"
Southern_Ndebele = "nr"
Spanish = "es"
Sundanese = "su"
Swahili_Kiswahili = "sw"
Swati = "ss"
Swedish = "sv"
Tagalog = "tl"
Tahitian = "ty"
Tajik = "tg"
Tamil = "ta"
Tatar = "tt"
Telugu = "te"
Thai = "th"
Tibetan = "bo"
Tigrinya = "ti"
Tonga = "to"
Tsonga = "ts"
Turkish = "tr"
Turkmen = "tk"
Twi = "tw"
Uyghur = "ug"
Ukrainian = "uk"
Urdu = "ur"
Uzbek = "uz"
Venda = "ve"
Vietnamese = "vi"
Volap = "vo"
Wallon = "wa"
Welsh = "cy"
Wolof = "wo"
Western_Frisian = "fy"
Xhosa = "xh"
Yiddish = "yi ji"
Yoruba = "yo"
Zhuang_Chuang = "za"
Zulu = "zu"



"""
    Color Names Supported by All Browsers
All modern browsers support the following 140 color names
(click on a color name, or a hex value, to view the color as the background-color along with different text colors):

    Returns:
        str: Color Hex Code

    functions:
        rgb_to_hex((r,g,b))  --> #RRGGBB
        hex_to_rgb(hex) --> (r,g,b)

"""

AliceBlue = "#F0F8FF"
AntiqueWhite = "#FAEBD7"
Aqua = "#00FFFF"
Aquamarine = "#7FFFD4"
Azure = "#F0FFFF"
Beige = "#F5F5DC"
Bisque = "#FFE4C4"
Black = "#000000"
BlanchedAlmond = "#FFEBCD"
Blue = "#0000FF"
BlueViolet = "#8A2BE2"
Brown = "#A52A2A"
BurlyWood = "#DEB887"
CadetBlue = "#5F9EA0"
Chartreuse = "#7FFF00"
Chocolate = "#D2691E"
Coral = "#FF7F50"
CornflowerBlue = "#6495ED"
Cornsilk = "#FFF8DC"
Crimson = "#DC143C"
Cyan = "#00FFFF"
DarkBlue = "#00008B"
DarkCyan = "#008B8B"
DarkGoldenRod = "#B8860B"
DarkGray = "#A9A9A9"
DarkGrey = "#A9A9A9"
DarkGreen = "#006400"
DarkKhaki = "#BDB76B"
DarkMagenta = "#8B008B"
DarkOliveGreen = "#556B2F"
DarkOrange = "#FF8C00"
DarkOrchid = "#9932CC"
DarkRed = "#8B0000"
DarkSalmon = "#E9967A"
DarkSeaGreen = "#8FBC8F"
DarkSlateBlue = "#483D8B"
DarkSlateGray = "#2F4F4F"
DarkSlateGrey = "#2F4F4F"
DarkTurquoise = "#00CED1"
DarkViolet = "#9400D3"
DeepPink = "#FF1493"
DeepSkyBlue = "#00BFFF"
DimGray = "#696969"
DimGrey = "#696969"
DodgerBlue = "#1E90FF"
FireBrick = "#B22222"
FloralWhite = "#FFFAF0"
ForestGreen = "#228B22"
Fuchsia = "#FF00FF"
Gainsboro = "#DCDCDC"
GhostWhite = "#F8F8FF"
Gold = "#FFD700"
GoldenRod = "#DAA520"
Gray = "#808080"
Grey = "#808080"
Green = "#008000"
GreenYellow = "#ADFF2F"
HoneyDew = "#F0FFF0"
HotPink = "#FF69B4"
IndianRed = "#CD5C5C"
Indigo = "#4B0082"
Ivory = "#FFFFF0"
Khaki = "#F0E68C"
Lavender = "#E6E6FA"
LavenderBlush = "#FFF0F5"
LawnGreen = "#7CFC00"
LemonChiffon = "#FFFACD"
LightBlue = "#ADD8E6"
LightCoral = "#F08080"
LightCyan = "#E0FFFF"
LightGoldenRodYellow = "#FAFAD2"
LightGray = "#D3D3D3"
LightGrey = "#D3D3D3"
LightGreen = "#90EE90"
LightPink = "#FFB6C1"
LightSalmon = "#FFA07A"
LightSeaGreen = "#20B2AA"
LightSkyBlue = "#87CEFA"
LightSlateGray = "#778899"
LightSlateGrey = "#778899"
LightSteelBlue = "#B0C4DE"
LightYellow = "#FFFFE0"
Lime = "#00FF00"
LimeGreen = "#32CD32"
Linen = "#FAF0E6"
Magenta = "#FF00FF"
Maroon = "#800000"
MediumAquaMarine = "#66CDAA"
MediumBlue = "#0000CD"
MediumOrchid = "#BA55D3"
MediumPurple = "#9370DB"
MediumSeaGreen = "#3CB371"
MediumSlateBlue = "#7B68EE"
MediumSpringGreen = "#00FA9A"
MediumTurquoise = "#48D1CC"
MediumVioletRed = "#C71585"
MidnightBlue = "#191970"
MintCream = "#F5FFFA"
MistyRose = "#FFE4E1"
Moccasin = "#FFE4B5"
NavajoWhite = "#FFDEAD"
Navy = "#000080"
OldLace = "#FDF5E6"
Olive = "#808000"
OliveDrab = "#6B8E23"
Orange = "#FFA500"
OrangeRed = "#FF4500"
Orchid = "#DA70D6"
PaleGoldenRod = "#EEE8AA"
PaleGreen = "#98FB98"
PaleTurquoise = "#AFEEEE"
PaleVioletRed = "#DB7093"
PapayaWhip = "#FFEFD5"
PeachPuff = "#FFDAB9"
Peru = "#CD853F"
Pink = "#FFC0CB"
Plum = "#DDA0DD"
PowderBlue = "#B0E0E6"
Purple = "#800080"
RebeccaPurple = "#663399"
Red = "#FF0000"
RosyBrown = "#BC8F8F"
RoyalBlue = "#4169E1"
SaddleBrown = "#8B4513"
Salmon = "#FA8072"
SandyBrown = "#F4A460"
SeaGreen = "#2E8B57"
SeaShell = "#FFF5EE"
Sienna = "#A0522D"
Silver = "#C0C0C0"
SkyBlue = "#87CEEB"
SlateBlue = "#6A5ACD"
SlateGray = "#708090"
SlateGrey = "#708090"
Snow = "#FFFAFA"
SpringGreen = "#00FF7F"
SteelBlue = "#4682B4"
Tan = "#D2B48C"
Teal = "#008080"
Thistle = "#D8BFD8"
Tomato = "#FF6347"
Turquoise = "#40E0D0"
Violet = "#EE82EE"
Wheat = "#F5DEB3"
White = "#FFFFFF"
WhiteSmoke = "#F5F5F5"
Yellow = "#FFFF00"
YellowGreen = "#9ACD32"


def rgb_to_hex(rgb_tuple):
    """
    Converts an RGB color represented as a tuple (r, g, b) into a hex color code.
    """
    hex_color = "#{:02X}{:02X}{:02X}".format(*rgb_tuple)
    return hex_color


def hex_to_rgb(hex_color):
    """
    Converts a hex color code into an RGB color represented as a tuple (r, g, b).
    """
    hex_color = hex_color.lstrip("#")
    r = int(hex_color[0:2], 16)
    g = int(hex_color[2:4], 16)
    b = int(hex_color[4:6], 16)
    return r, g, b



import argparse
import re

try:
    from lxml import etree
except ImportError:
    raise "you must install lxml, run `python3 -m pip install lxml"
try:

    from black import FileMode, format_str
except ImportError:
    raise "you must install black, run `python3 -m pip install black"
import inui
import inui.base
import inui.bootstrap
import inui.colors
import inui.css
import inui.elements
import inui.svg
import inui.tailwind
from inui.config.replaces import reverse_replace

create_inui_template = """

#* This Element Not Found in inui project, we will create an equivalent element for you."

from inui.base import BaseElement, _Meta

class %s(BaseElement):
    _Meta = _Meta("%s", "%s")


#* -------------------------------------------------------------------------------------
"""


class Element:
    def __init__(self, name="", children=None, attributes=None, data=None):
        self.name = name
        self.children = children if children is not None else []
        self.attributes = attributes if attributes is not None else {}
        self.data = data

    def add_child(self, child):
        self.children.append(child)

    def string_quote(self, value):
        if "\n" not in value:
            if "'" in value and '"' in value:
                #! handle both ''' and """ in text
                if '"""' not in value:
                    return f'"""{value}"""'
                elif "''''" not in value:
                    return f"''''{value}''''"
            elif "'" not in value:
                return f"'{value}'"
            elif '"' not in value:
                return f'"{value}"'
        return f"'''{value}'''"

    def __str__(self, indent=0):
        element_str = ""
        # if indent == 0:
        #     element_str += self
        attind = " " * (indent + 4)
        attributes_str = "\n".join(
            [
                f"{attind}{reverse_replace(attr)}={self.string_quote(value)},"
                for attr, value in self.attributes.items()
            ]
        )
        element_str += f"{' ' * indent}{self.name}(\n"

        # element_str = f"{' ' * indent}"
        if self.data or self.children:
            element_str += "\n" + attind
        if self.data:
            if self.data.strip():
                # print(self.data)
                element_str += f"{self.string_quote(self.data)}," if self.data else ""
        for child in self.children:
            if isinstance(child, str):
                # print(type(child))
                element_str += "\n'''" + child.__str__(indent + 8) + ", '''"
            else:
                element_str += "\n" + child.__str__(indent + 8) + ", "

        element_str += f"\n{' ' * indent}"
        element_str += f"{attributes_str}"
        element_str += ")"
        while "''''''" in element_str:
            element_str = element_str.replace("''''''", "''")
        while "''''" in element_str:
            element_str = element_str.replace("''''", "'''")
        return element_str


def create_element_tree(html_text):
    root = etree.HTML(html_text)
    return create_element_from_etree(root)


def create_element_from_etree(element: etree._Element):
    attributes = {}
    if element.attrib:
        attributes = element.attrib

    children = []
    for child in element:
        children.append(create_element_from_etree(child))

    data = element.text if element.text else None

    return Element(
        name=reverse_replace(element.tag).capitalize(),
        children=children,
        attributes=attributes,
        data=data,
    )


class HtmlTo_inui:
    def __init__(
        self, html=None, url=None, fileName=None, saveTo=None, indent=4
    ) -> None:
        self.html = html
        self.url = url
        self.filename = fileName
        self.indent = indent
        self.inui = None
        self.imports = []
        self.not_imported = []

    def to_inui(self, html_text=None):
        root = etree.HTML(self._remove_html_comments(html_text or self.html or ""))
        return self.add_imports(self._create_element_from_etree(root))

    def add_imports(self, code: str):
        imports = self.manage_imports(self.imports)
        not_imported = self.create_unknown_elements()
        return self.format(
            imports
            + not_imported
            + "content = "
            + str(code)
            + "\n\n #? run via: `inui file_name:content`"
        )

    def format(self, code):
        try:
            return format_str(str(code), mode=FileMode())
        except Exception as _:
            return str(code)

    def _remove_html_comments(self, html):
        pattern = r"<!--(.*?)-->"
        clean_html = re.sub(pattern, "", html, flags=re.DOTALL)
        return clean_html

    def manage_imports(self, imports):
        places = {
            "base": [],
            "elements": [],
            "svg": [],
            "tailwind": [],
            "bootstrap": [],
            "css": [],
            "colors": [],
        }
        for imp in set(imports):
            imported = False
            for place in places:
                if imported:
                    continue
                if hasattr(getattr(inui, place), imp):
                    places[place].append(imp)
                    imported = True

            if not imported:
                self.not_imported.append(imp)

        res = ""
        for module, item in places.items():
            if item:
                res += f"from inui.{module} import {self.print_list(sorted(item))}\n"
        return res

    def print_list(self, data: tuple):
        return (
            "(" + ", ".join([str(i) for i in data]) + ")"
            if len(data) > 1
            else str(data[0]) if len(data) > 0 else ""
        )

    def _create_element_from_etree(self, element: etree._Element):
        self.imports.append(reverse_replace(element.tag).capitalize())
        attributes = {}
        if element.attrib:
            attributes = element.attrib

        children = []
        for child in element:
            children.append(self._create_element_from_etree(child))

        data = element.text if element.text else None

        self.inui = Element(
            name=reverse_replace(element.tag).capitalize(),
            children=children,
            attributes=attributes,
            data=data,
        )
        return self.inui

    def _create_element(self, name: str):
        return create_inui_template % (reverse_replace(name).capitalize(), name, name)

    def create_unknown_elements(self):
        res = ""
        for i in self.not_imported:
            res += self._create_element(i) + "\n"

        return res

    def from_url(self, url):
        try:
            import requests
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                'requests package not found, try: "python3 -m pip install requests"'
            )
        self.url = url
        self.html = requests.get(self.url).text
        return self.to_inui(self.html)

    def file(self, path=None):
        if path is not None:
            self.filename = path
        with open(self.filename, "r", encoding="utf-8") as f:
            self.html = f.read()
        return self.to_inui(self.html)

    def save(self, path):
        if self.inui:
            self.to_inui(self.html)
        with open(path, "w", encoding="utf-8") as f:
            f.write(str(self.inui))
        return True


def main():
    parser = argparse.ArgumentParser(description="Convert HTML to inui format")
    parser.add_argument("--html", type=str, help="HTML content as a string")
    parser.add_argument("--url", type=str, help="URL to fetch HTML content from")
    parser.add_argument("--fileName", type=str, help="Path to the HTML file")
    parser.add_argument("--saveTo", type=str, help="Path to save the output file")
    parser.add_argument(
        "--indent", type=int, default=4, help="Indentation level for output"
    )

    args = parser.parse_args()

    converter = HtmlTo_inui(
        html=args.html,
        url=args.url,
        fileName=args.fileName,
        saveTo=args.saveTo,
        indent=args.indent,
    )

    if args.url:
        output = converter.from_url(args.url)
    elif args.fileName:
        output = converter.file(args.fileName)
    elif args.html:
        output = converter.to_inui(args.html)
    else:
        raise ValueError("Either --html, --url or --fileName must be provided")

    if args.saveTo:
        converter.save(args.saveTo)
        print(f"Output saved to {args.saveTo}")
    else:
        print(output)


if __name__ == "__main__":
    main()
